// ***********************************************************
// Project:  Power-Line-Modul mit ATtiny24
// Software: AVR-Studio v4.18 / WinAVR-20100110
// Author:   Jens Grtler
// Erstellt: 24.11.2011
// ***********************************************************


// --- einbinden von Libs ------------------------------------
#include <avr\io.h>                          // AVR Basisdaten einbinden
#include <avr/interrupt.h>                   // Interrupts bereitstellen
#include <avr/eeprom.h>                      // EEProm zugriff bereitstellen
#include <avr/wdt.h>                         // Watchdog Routine einbinden

// --- User Konstanten definieren ----------------------------
#define Ladezeit         24                  // Ladezeit = 24 * 5 Min = 2 Stunden
#define Lade_Wert        365                 // Ladeschwelle (ca.= 3,7V)

// --- Definiton der Konstanten ------------------------------
#define Tim1_vorladen    63036               // Vorladen 20ms
#define Time_10s         500                 // 500 * 20ms = 10 Sekunden

// --- Definiton E-Eprom Adressen ----------------------------
uint8_t Dummi_EEprom EEMEM = 0;              // Dummi, wird nicht verwendet
uint8_t Akku_EEprom EEMEM = 0;               // Ladezeit Akku

// --- Definiton der Globalen Variabeln ----------------------
volatile uint8_t Zeit = 0;                   // Zeit Zhler
volatile uint8_t Power_ON = 130;             // Zeit Zhler fr Power-ON
volatile uint16_t Power_OFF = Time_10s;      // Zeit Zhler einstellen


// *************************************************************************************
// *   INTERRUPT Vektor 8 - Timer1 berlauf 20ms                                       *
// *************************************************************************************
ISR(TIM1_OVF_vect) {                         // Beginn Timer1 berlauf
   TCNT1 = Tim1_vorladen;                    // Timer1 neu vorladen
   if (Power_ON > 0) { Power_ON--; }         // Zeit Zhler -1
   if (Power_OFF > 0) { Power_OFF--; }       // Zeit Zhler -1
   Zeit++;                                   // Zeit Zhler +1
}                                            // Ende Timer1 berlauf


// *************************************************************************************
// *   FUNKTION - AD-Wert einlesen                                                     *
// *************************************************************************************
uint16_t ADC_einlesen(uint8_t Kanal) {       // AD-Wert einlesen
   uint16_t Ergebnis = 0;                    // Ergebnis AD-Messung
   // Kanal einstellen, ohne andere Bits zu beeinfluen
   ADMUX = (ADMUX & ~(0x1F)) | (Kanal & 0x1F);  // Kanal einstellen
   // Wandlermessung - verwerfen
   ADCSRA |= (1<<ADSC);                      // Messung starten
   while (ADCSRA & (1<<ADSC) ) { asm volatile ("nop"); }  // auf Abschluss der Konvertierung warten
   // 4 Wandlermessungen
   for (uint8_t i = 0; i < 8; ++i ) {        // For 0 To 8 Step 1
      ADCSRA |= (1<<ADSC);                   // Messung starten
      while (ADCSRA & (1<<ADSC) ) { asm volatile ("nop"); }  // auf Abschluss der Konvertierung warten
      Ergebnis += ADCW;                      // Messung zwischenspeichern
   }                                         // Next 
   // Ergebnis durch 8 Teilen
   Ergebnis = Ergebnis / 8;                  // Mittelwert berechnen
   // bergabe
   return Ergebnis;                          // Messung zurckgeben
}                                            // Ende AD-Wert einlesen


// *************************************************************************************
// *   FUNKTION - Akku berprfen & laden                                              *
// *************************************************************************************
void Akku_ueberpruefen(uint8_t Sta) {        // Akku berprfen
   uint8_t Betrieb = eeprom_read_byte(&Akku_EEprom);  // Akku Ladezeit einlesen
   uint16_t Akku_Min_Wert = ADC_einlesen(3);  // Z-Diode einlesen
   uint16_t Akku_Spannung = ADC_einlesen(1);  // Akku Spannung einlesen
   // Sicherheit - EEprom Daten falsch
   if (Betrieb > Ladezeit) {                 // Sicherheit - MAX Wert abfragen
      Betrieb = Ladezeit;                    // Ladezeit einstellen
      eeprom_write_byte(&Akku_EEprom, Betrieb);  // Ladezeit in EEprom schreiben
   }                                         // Ende Sicherheit - MAX Wert abfragen
   // Ladevorgang nicht aktiv
   if (Betrieb == 0) {                       // wenn kein Ladevorgang luft, dann...
      Akku_Min_Wert = Akku_Min_Wert + Lade_Wert;  // Z-Diode + Min-Wert berechnen
      if (Akku_Spannung <= Akku_Min_Wert) {  // wenn Akku unter 3,7V, dann...
         Betrieb = Ladezeit;                 // Ladezeit einstellen
         eeprom_write_byte(&Akku_EEprom, Betrieb);  // Ladezeit in EEprom schreiben
      }                                      // Ende wenn Akku unter 3,7V
   }                                         // Ende wenn kein Ladevorgang luft
   if (Sta > 0) {                            // wenn laufender Betrieb, dann...
      // Ladevorgang aktiv
      if (Betrieb > 0) {                     // Ladevorgang luft
         Betrieb--;                          // Ladezeit -1
         eeprom_write_byte(&Akku_EEprom, Betrieb);  // restliche Ladezeit in EEprom schreiben
      }                                      // Ende Ladevorgang luft
   }                                         // Ende wenn laufender Betrieb
   // Lade-Port schalten
   if (Betrieb == 0) {                       // wenn inaktiv, dann...
      PORTA &= ~(1<<PA2);                    // Load-Enabel AUSschalten
      PORTA &= ~(1<<PA5);                    // Test Pin AUSschalten
   }                                         // Ende wenn inaktiv
   else {                                    // wenn aktiv, dann...
      PORTA |= (1<<PA2);                     // Load-Enabel EINschalten
      PORTA |= (1<<PA5);                     // Test Pin EINschalten
   }                                         // Ende wenn aktiv
}                                            // Ende Akku berprfen


// *************************************************************************************
// *   Hauptprogramm                                                                   *
// *************************************************************************************
int main(void) {                             // Beginn Hauptprogramm

   // --- Watchdog abschalten (nach einen Reset) ---------------------------------------
   MCUSR &= ~(1<<WDRF);                      // Watchdog OFF
   WDTCSR |= (1<<WDCE) | (1<<WDE);           // Watchdog OFF
   WDTCSR = 0x00;                            // Watchdog OFF

   // --- Definiton der Lokalen Variabeln ----------------------------------------------
   uint16_t Tmp01 = 0;                       // Temp Variable Nr.1
   uint8_t Sekunde = 0;                      // Sekunde
   uint8_t Min = 0;                          // Minuten Schleife
   uint8_t ON_Flag = 0;                      // PC Netzteil AN Flag
   uint8_t Prg_Start = 0;                    // Zeit Zhler Start

   // --- Eingnge setzen --------------------------------------------------------------
   // PB0   Frei
   // PB1   Frei
   // PB3   Reset
   DDRB &= ~( (1<<DDB0) | (1<<DDB1) );       // B-Register - PB0 und PB1 als Eingang setzen
   PORTB |= (1<<PB0) | (1<<PB1);             // Interne Pull-Up setzen (1)
   // PA7   ADC - 5V vom PC-Netzteil
   // PA6   ADC - Power-Taster (MOSI)
   // PA3   ADC - Z-Diode 3,7V
   // PA1   ADC - Akku Spannung (wenn Relais ein und Load aus)
   DDRA &= ~( (1<<DDA1) | (1<<DDA3) | (1<<DDA6) | (1<<DDA7) );  // A-Register - PA1, PA3, PA6 und PA7 als Eingang setzen
   PORTA &= ~( (1<<PA1) | (1<<PA6) | (1<<PA7) );                // Interne Pull-Up abschalten (0)
   PORTA |= (1<<PA3);                        // Interne Pull-Up einschalten (1)


   // --- Ausgnge setzen --------------------------------------------------------------
   // PB2   MoBo-Power-ON (2x FET)
   DDRB = (1<<DDB2);                         // B-Register - PB2 als Ausgang setzen
   PORTB &= ~(1<<PB2);                       // Port ausschalten (0)
   // PA5   Test Pin - Frei (MISO) - LED Ladezeit
   // PA4   Test Pin - Frei (SCK)
   // PA2   Load-Enabel - MAX1811 (Masse = Aus)
   // PA0   Relais Akku
   DDRA = (1<<DDA0) | (1<<DDA2) | (1<<DDA4) | (1<<DDA5);  // D-Register - PA0, PA2, PA4 und PA5 als Ausgang setzen
   PORTA &= ~( (1<<PA0) | (1<<PA2) | (1<<PA4) | (1<<PA5) );  // PA0, PA2, PA4 und PA5 - Ports ausschalten (0)

   // --- AD-Wandler aktivieren --------------------------------------------------------
   ADCSRA = (1<<ADEN) | (1<<ADPS2);          // AD-Wandler aktivieren, Frequenzvorteiler 16

   // --- Timer1 - 20ms ----------------------------------------------------------------
   TCCR1B = (1<<CS11);                       // Vorteiler 8
   TIMSK1 = (1<<TOIE1);                      // Overflow Interrupt Enable
   TCNT1 = Tim1_vorladen;                    // Timer1 vorladen

   // --- Interrupt aktivieren ---------------------------------------------------------
   sei();                                    // Interrupts aktivieren

   // --- Voreinstellung ---------------------------------------------------------------
   PORTA |= (1<<PA0);                        // Relais Akku EINschalten

   // --- Watchdog aktivieren ----------------------------------------------------------
   wdt_enable(WDTO_1S);                      // Watchdog auf 1 Sekunde einstellen


   // **********************************************************************************
   // *   Endlosschleife                                                               *
   // **********************************************************************************
   while(1) {                                // Beginn While - Endlosschleife

      // --- 10 Sekunden Power-OFF -----------------------------------------------------
      Tmp01 = ADC_einlesen(7);               // AD-Wert vom PC-Netzteil einlesen
      if (Tmp01 >= 300) {                    // wenn PC AN, dann...
         Power_OFF = Time_10s;               // Reset - Zeit Zhler zurck setzen
         ON_Flag = 1;                        // PC AN
      }                                      // Ende wenn PC AN
      else { ON_Flag = 0; }                  // PC AUS
      if (Power_OFF == 0) { PORTA &= ~(1<<PA0); }  // Relais Akku AUSschalten

      // --- Taster abfragen -----------------------------------------------------------
      Tmp01 = ADC_einlesen(6);               // Taster einlesen
      if (Tmp01 >= 200) {                    // wenn Taster bettigt, dann...
         Power_OFF = Time_10s;               // Reset - Zeit Zhler zurck setzen
         if (Prg_Start > 10) {               // 10 Sek nach Programmstart
            if (ON_Flag == 1) {              // wenn PC AN, dann...
               Power_ON = 5;                 // Mobo einschalten
            }                                // Ende wenn PC AN
            else {                           // wenn PC AUS, dann...
               Prg_Start = 0;                // Reset - Programmstart
               PORTA |= (1<<PA0);            // Relais Akku EINschalten
               Power_ON = 130;               // Reset - Zeit voreinstellen
            }                                // Ende wenn PC AUS
         }                                   // Ende 10 Sek nach Programmstart
      }                                      // Ende wenn Taster bettigt

      // --- Power Mobo AN -------------------------------------------------------------
      if ((Power_ON > 0) && (Power_ON <= 30)) { PORTB |= (1<<PB2); }  // Power-Mobo ON
      else { PORTB &= ~(1<<PB2); }           // Power-Mobo OFF

      // --- Zeit verarbeiten ----------------------------------------------------------
      if (Zeit >= 50) {                      // wenn 1 Sekunde vorbei, dann...
         Zeit = 0;                           // Reset Zeit Zhler
         // *** System Start Zhler ***
         if (Prg_Start < 20) { Prg_Start++; }  // Programmstart mit zhlen
         // *** System Start - nach 8 Sekunden ***
         if ((Prg_Start == 8) && (ON_Flag == 1)) {  // nach 8 Sekunden ausfhren
            Akku_ueberpruefen(0);            // Akkuspannung berprfen & Starten
         }                                   // Ende nach 8 Sekunden ausfhren
         // Sekunde
         Sekunde++;                          // Sekunde +1
         if (Sekunde > 59) {                 // wenn 1 Minute vorbei, dann...
            Sekunde = 0;                     // Reset - zurck setzen
            // Minute
            Min++;                           // Minute +1
         }                                   // Ende wenn 1 Minute vorbei
      }                                      // Ende wenn 1 Sekunde vorbei

      // --- 5 Minuten Zeitschleife - Akku verwalten -----------------------------------
      if (Min >= 5) {                        // wenn 5 Minuten vorbei, dann...
         Min = 0;                            // Reset Zeit Zhler
         if (ON_Flag == 1) {                 // wenn PC AN ist, dann...
            Akku_ueberpruefen(1);            // Akkuspannung berprfen & laden
         }                                   // Ende wenn PC AN ist
      }                                      // Ende wenn 5 Minuten vorbei

      // --- Hardware Watchdog zurcksetzen --------------------------------------------
      wdt_reset();                           // Watchdog Reset

   }                                         // Ende While
}                                            // Ende Hauptprogramm


//    PORTB |= (1<<PB0);                     // Bussy-LED EINschalten
//    PORTB ^= (1<<PB0);                     // Toggle Bussy-LED
//    PORTB &= ~(1<<PB0);                    // Bussy-LED AUSschalten






